@@ -1,6 +1,7 @@
-import {useData} from 'argo-ui/v2';
+import {useData, Checkbox} from 'argo-ui/v2';
 import * as minimatch from 'minimatch';
 import * as React from 'react';
+import {Context} from '../../../shared/context';
 import {Application, ApplicationDestination, Cluster, HealthStatusCode, HealthStatuses, SyncStatusCode, SyncStatuses} from '../../../shared/models';
 import {AppsListPreferences, services} from '../../../shared/services';
 import {Filter, FiltersGroup} from '../filter/filter';
@@ -8,12 +9,12 @@ import * as LabelSelector from '../label-selector';
 import {ComparisonStatusIcon, HealthStatusIcon} from '../utils';
 
 export interface FilterResult {
-    projects: boolean;
     repos: boolean;
     sync: boolean;
     health: boolean;
     namespaces: boolean;
     clusters: boolean;
+    favourite: boolean;
     labels: boolean;
 }
 
@@ -25,11 +26,11 @@ export function getFilterResults(applications: Application[], pref: AppsListPref
     return applications.map(app => ({
         ...app,
         filterResult: {
-            projects: pref.projectsFilter.length === 0 || pref.projectsFilter.includes(app.spec.project),
             repos: pref.reposFilter.length === 0 || pref.reposFilter.includes(app.spec.source.repoURL),
             sync: pref.syncFilter.length === 0 || pref.syncFilter.includes(app.status.sync.status),
             health: pref.healthFilter.length === 0 || pref.healthFilter.includes(app.status.health.status),
             namespaces: pref.namespacesFilter.length === 0 || pref.namespacesFilter.some(ns => app.spec.destination.namespace && minimatch(app.spec.destination.namespace, ns)),
+            favourite: !pref.showFavorites || (pref.favoritesAppList && pref.favoritesAppList.includes(app.metadata.name)),
             clusters:
                 pref.clustersFilter.length === 0 ||
                 pref.clustersFilter.some(filterString => {
@@ -60,6 +61,7 @@ interface AppFilterProps {
     pref: AppsListPreferences;
     onChange: (newPrefs: AppsListPreferences) => void;
     children?: React.ReactNode;
+    collapsed?: boolean;
 }
 
 const getCounts = (apps: FilteredApp[], filterType: keyof FilterResult, filter: (app: Application) => string, init?: string[]) => {
@@ -213,13 +215,36 @@ const NamespaceFilter = (props: AppFilterProps) => {
     );
 };
 
-export const ApplicationsFilter = (props: AppFilterProps) => {
-    const setShown = (val: boolean) => {
-        services.viewPreferences.updatePreferences({appList: {...props.pref, hideFilters: !val}});
+const FavoriteFilter = (props: AppFilterProps) => {
+    const ctx = React.useContext(Context);
+    const onChange = (val: boolean) => {
+        ctx.navigation.goto('.', {showFavorites: val}, {replace: true});
+        services.viewPreferences.updatePreferences({appList: {...props.pref, showFavorites: val}});
     };
-
     return (
-        <FiltersGroup setShown={setShown} expanded={!props.pref.hideFilters} content={props.children}>
+        <div
+            className={`filter filter__item ${props.pref.showFavorites ? 'filter__item--selected' : ''}`}
+            style={{margin: '0.5em 0'}}
+            onClick={() => onChange(!props.pref.showFavorites)}>
+            <Checkbox
+                value={!!props.pref.showFavorites}
+                onChange={onChange}
+                style={{
+                    marginRight: '8px'
+                }}
+            />
+            <div style={{marginRight: '5px'}}>
+                <i style={{color: '#FFCE25'}} className='fas fa-star' />
+            </div>
+            <div className='filter__item__label'>Favorites Only</div>
+        </div>
+    );
+};
+
+export const ApplicationsFilter = (props: AppFilterProps) => {
+    return (
+        <FiltersGroup content={props.children} collapsed={props.collapsed}>
+            <FavoriteFilter {...props} />
             <SyncFilter {...props} />
             <HealthFilter {...props} />
             <LabelsFilter {...props} />
