@@ -14,7 +14,7 @@ func Test_cmd_redactor(t *testing.T) {
 }
 
 func TestCmd_template_kubeVersion(t *testing.T) {
-	cmd, err := NewCmdWithVersion(".", HelmV3, false)
+	cmd, err := NewCmdWithVersion(".", HelmV3, false, "")
 	assert.NoError(t, err)
 	s, err := cmd.template("testdata/redis", &TemplateOpts{
 		KubeVersion: "1.14",
@@ -23,26 +23,26 @@ func TestCmd_template_kubeVersion(t *testing.T) {
 	assert.NotEmpty(t, s)
 }
 
-func TestNewCmd_helmV2(t *testing.T) {
-	cmd, err := NewCmd(".", "v2")
-	assert.NoError(t, err)
-	assert.Equal(t, "helm2", cmd.HelmVer.binaryName)
-}
-
 func TestNewCmd_helmV3(t *testing.T) {
-	cmd, err := NewCmd(".", "v3")
+	cmd, err := NewCmd(".", "v3", "")
 	assert.NoError(t, err)
 	assert.Equal(t, "helm", cmd.HelmVer.binaryName)
 }
 
 func TestNewCmd_helmDefaultVersion(t *testing.T) {
-	cmd, err := NewCmd(".", "")
+	cmd, err := NewCmd(".", "", "")
 	assert.NoError(t, err)
 	assert.Equal(t, "helm", cmd.HelmVer.binaryName)
 }
 
 func TestNewCmd_helmInvalidVersion(t *testing.T) {
-	_, err := NewCmd(".", "abcd")
+	_, err := NewCmd(".", "abcd", "")
 	log.Println(err)
 	assert.EqualError(t, err, "helm chart version 'abcd' is not supported")
 }
+
+func TestNewCmd_withProxy(t *testing.T) {
+	cmd, err := NewCmd(".", "", "https://proxy:8888")
+	assert.NoError(t, err)
+	assert.Equal(t, "https://proxy:8888", cmd.proxy)
+}
